#!/usr/bin/perl -w

# EnergyProfiler v1.0
# Tony Flick (tony.flick[at)fyrmassociates.com)
# Justin Morehouse (justin.morehouse(at]gmail.com)

# This script provides a visual profile of a stalkee's energy usage. 
# It assumes that stalkees_data.pl has already been run and created
# some data files.

# The latest version is always available for download at http://www.fyrmassociates.com/tools.html

use CGI ':standard';
use HTML::Entities;
use Tie::File;
use GD::Graph::bars;
use GD::Graph::linespoints;
use Time::Piece;

my ($q, $stalkee, $daily, $tzoffset, $chartTitle, @times, @watts, @chartData);


$q = new CGI();

# If the stalkee parameter exists
if ($q->param("stalkee")){
	$stalkee = $q->param("stalkee");
	$fileName = "DataFiles/" . $stalkee . ".csv";

	# If the file exists
	if (-e $fileName){
		tie @fileLines, 'Tie::File', $fileName;
		my @tempTZ = parseCSV($fileLines[0]);
		$tzoffset = $tempTZ[scalar(@tempTZ) - 2];

		# if the daily parameter exists, then we will either print out the 
		# average daily usage or a specific day's usage
		if ($q->param("daily")){
			$daily = $q->param("daily");
		
			# Print out the daily average for each hour of the day for whatever data we have
			if ($daily eq "average"){
				# Hash of each hour's average
				#Keys: 0 = 12:00am, 1 = 1:00am, 13 = 1:00pm (13:00)
				my %hourHash;
			}

			# Print out a specific day's usage
			# Expected input format is mmddyyyy or sun,mon,tue,wed,thu,fri,sat
			else{
				# If they entered a spcific date: mm-dd-yyyy
				if ($daily =~ m/^\d\d-\d\d-\d\d\d\d$/){
					my $inTime = Time::Piece->strptime($daily, "%m-%d-%Y");
					$chartTitle = "Wattage For " . $inTime->ymd();

					foreach my $line (@fileLines){
						my @thisTweet = parseTweets($line);
						my $tweetTime = Time::Piece->strptime($thisTweet[1], "%a %b %d %H:%M:%S %z %Y");
						# Multiply the offset (number of hours) by the number of seconds in an hour
						$tweetTime += $tzoffset;

						if ($inTime->ymd() eq $tweetTime->ymd()){
							unshift(@watts, $thisTweet[0]);
							unshift(@times, $tweetTime->hms());
						}
					}
				}

				# If they want a specific day of the week
				elsif (($daily =~ m/^sunday$/i) || ($daily =~ m/^monday$/i) || ($daily =~ m/^tuesday$/i) ||
					($daily =~ m/^wednesday$/i) || ($daily =~ m/^thursday$/i) || 
					($daily =~ m/^friday$/i) || ($daily =~ m/^saturday$/i)){
					my $dailyShort = substr($daily, 0, 3);
					# Stores the average equations for each hour
					# index: 0-23 is the hour time
					my @timeWatts = ("0)/1") x 24;

					# Read through each line of the data file looking for the specific day
					foreach my $line (@fileLines){
						my @thisTweet = parseTweets($line);
						my $tweetTime = Time::Piece->strptime($thisTweet[1], "%a %b %d %H:%M:%S %z %Y");

						# Use the timezone offset to calculate the real day
						$tweetTime += $tzoffset;					
						
						my $tweetDayTemp = $tweetTime->wdayname();
						# if the input day matches that particular tweets day
						if ($dailyShort =~ m/^$tweetDayTemp$/i){
							$chartTitle = "Average Wattage For " . $tweetTime->fullday() . "s";
	
							my $hourIndex = $tweetTime->hour();
							my $tempDivisor = substr($timeWatts[$hourIndex], -1);
							$tempDivisor += 1;
							$timeWatts[$hourIndex] = $thisTweet[0] . " + " . substr($timeWatts[$hourIndex], 0, -1) . $tempDivisor;
						}
					}

					my $iCount = 0;
					# Assign the hours to the time array for the y-axis
					while ($iCount < 24){
						if ($iCount < 10){
							$times[$iCount] = "0" . $iCount . ":00";
						}
						else{
							$times[$iCount] = $iCount . ":00";
						}
						$iCount += 1;
					}

					# Calculate the averages for each hour
					$hourCount = 0;
					@watts = (0..23);
					while ($hourCount < scalar(@timeWatts)){
						my $tempDivisor2 = substr($timeWatts[$hourCount], -1);
						if ($tempDivisor2 > 1){
							$tempDivisor2 -= 1;
						}
						my $thisHour = "(" . substr($timeWatts[$hourCount], 0, -1) . $tempDivisor2;
						$watts[$hourCount] = eval($thisHour);
						$hourCount += 1;
					}

				}

				else{
					print "Content-type: text/html\n\n";
					print "Please enter a correct date format.<p>";
					print "The daily parameter takes the following values.<p>Specific date: mm-dd-yyyy";
					print "<p>For a specific day of the week: sun, mon, tue, wed, thu, fri, sat";
				}

			}

		} # ends if ($q->param("daily"))


		# chart all data
		elsif ($q->param("all")){
			# Parse each tweet for the watts and date
			foreach my $line (@fileLines){
				my @thisTweet = parseTweets($line);
				unshift(@watts, $thisTweet[0]);

				my $tweetTime = Time::Piece->strptime($thisTweet[1], "%a %b %d %H:%M:%S %z %Y");
				unshift(@times, $tweetTime->mdy() . " " . $tweetTime->hms());
				$chartTitle = "All Data Availaable";
			}
		}

		else {
			print "Content-type: text/html\n\n";
			print "Not enough parameters";
		}
		
		untie @fileLines;



		if (scalar(@watts) == 0){
			print "Content-type: text/html\n\n";
			print "No data was found for the requested timeframe.";
		}
		else{
			printChart($chartTitle, \@watts, \@times);
		}
	} # ends if file exists block

	else{
		print "Content-type: text/html\n\n";
		print HTML::Entities::encode($stalkee), " is not being followed yet.<p>";
		print "Please use stalkees_data.pl to start stalking an account first.<p>";
	}
}

else{		
	print "Content-type: text/html\n\n";
	print "Please use and enter a value for the stalkee parameter. ";
	print "Usage: energy_profiler.pl?stalkee=&lt;twitter name&gt;<br />";
}


# Print the chart for the specified data
sub printChart {
	my ($dateString, $dataArrayRef1, $dataArrayRef2) = @_;
	my @dataArray1 = @{$dataArrayRef1};
	my @dataArray2 = @{$dataArrayRef2};

	my $graph = GD::Graph::lines->new(120*scalar(@dataArray1), 400);
	$graph->set(
	    x_label     => 'Time',
	    y_label     => 'Watts',
	    title       => $dateString,
	) or warn $graph->error;


	

	open(OFILE, ">DataFiles/socawatt.txt");
	$count = 0;
	while ($count < scalar(@dataArray1)){
		print OFILE $dataArray1[$count] . ", " . $dataArray2[$count] . "\n";
		$count += 1;
	}
	close(OFILE);

	my @dataArray = (\@dataArray2, \@dataArray1);
	my $format = $graph->export_format;
	print header("image/$format");
	binmode STDOUT;
	print $graph->plot(\@dataArray)->$format();
}


# Parse a string that holds a single line from the stalkee data file
# Should return an array of the watts and date
# parsedTweet[0] = <watts>
# parsedTweets[1] = <time>
sub parseTweets {

	my $inText = shift;
	my @parsedTweet = ();
	my @parsedLine = parseCSV($inText);

	# Grab the Watt amounts
	my @tempWattage;
	if ($parsedLine[1] =~ m/Watt/){
		@tempWattage = split('Watt', $parsedLine[1]);
	}
	elsif ($parsedLine[1] =~ m/\sW,/){
		@tempWattage = split(/\sW/, $parsedLine[1]);
	}
	my @watt = split(' ', $tempWattage[0]);
	my $wattLength = scalar(@watt);
	push(@parsedTweet, $watt[$wattLength-1]);
	
	# Grab the date
	push(@parsedTweet, $parsedLine[2]);	

	return @parsedTweet;
}


# Parse a csv string into an array
sub parseCSV {
  my $text = shift;
  my @new = ();
  push (@new, $+) while $text =~ m{
          "([^\"\\]*(?:\\.[^\"\\]*)*)",?
          | ([^,]+),?
          | ,
      }gx;
  push (@new, undef) if substr($text, -1, 1) eq ",";
  return @new;
}
