#!/usr/bin/perl -w

# Stalker v1.0
# Tony Flick (tony.flick[at)fyrmassociates.com)
# Justin Morehouse (justin.morehouse(at]gmail.com)

# Predicts when a person is home and when they are sleeping

# The latest version is always available for download at http://www.fyrmassociates.com/tools.html

use CGI ':standard';
use HTML::Entities;
use Tie::File;
use GD::Graph::bars;
use GD::Graph::linespoints;
use Time::Piece;

use Data::ICal;
use Data::ICal::Entry::Event;
use Data::ICal::DateTime;
use Date::Manip;

# Constant for the error used to calculate deviations from the average lows
my $ERROR = 10;

# Constant for the starting address for google map directions
my $CURRENTADDRESS = "36.132,+-115.157";

my ($date, @prediction);

$q = new CGI();
print "Content-type: text/html\n\n";

# If the stalkee and date parameters exist, then we have all the parameters we need
if (($q->param("stalkee")) && ($q->param("date"))){
	$stalkee = $q->param("stalkee");
	$fileName = "DataFiles/" . $stalkee . ".csv";
	$date = $q->param("date");


	# If they want to see when everyone can be robbed
	if ($stalkee =~ m/^all$/i){
		my $inTime;
		my $dateString;
		my $timeIn;

		# If they entered a spcific date: mm-dd-yyyy
		# Predict based on day of the week
		if ($date =~ m/^\d\d-\d\d-\d\d\d\d$/){
			$inTime = Time::Piece->strptime($date, "%m-%d-%Y");
			$dateString = $inTime->mdy();
			$timeIn = inTime->wdayname();
		}
		else{
			$dateString = $date . "s";
			$timeIn = $date;
		}

		opendir(DIRECTORY, "DataFiles/");
		foreach my $file (readdir(DIRECTORY)){
			if ($file =~ m/.csv$/){
				my @splitFile = split(".csv", $file);
				predictWeekDay($splitFile[0], $dateString, $timeIn, "DataFiles/" . $file);
			}
		}
		closedir(DIRECTORY);
	}


	# If the file exists
	elsif (-e $fileName){
		# If they entered a spcific date: mm-dd-yyyy
		# Predict based on day of the week
		if ($date =~ m/^\d\d-\d\d-\d\d\d\d$/){
			my $inTime = Time::Piece->strptime($date, "%m-%d-%Y");
			my $dateString = $inTime->mdy();
			predictWeekDay($stalkee, $dateString, $inTime->wdayname(), $fileName);
		}

		elsif (($date =~ m/^sunday$/i) || ($date =~ m/^monday$/i) || ($date =~ m/^tuesday$/i) ||
			($date =~ m/^wednesday$/i) || ($date =~ m/^thursday$/i) || 
			($date =~ m/^friday$/i) || ($date =~ m/^saturday$/i)){
			my $dateString = $date . "s";
			predictWeekDay($stalkee, $dateString, $date, $fileName);
		}

		else {
			print "Usage: stalker.pl?stalkee=&lt;twitter_name&gt;&date=[mm-dd-yyyy || 7-letter weekday]";
			print "Example: stalker.pl?stalkee=tonysmartmeter&date=Thursday";
		}
	}

	else{
		print HTML::Entities::encode($q->param("stalkee")), " is not being stalked yet. Please use stalkees_data.pl first.";
	}

} # Ends if (($q->param("stalkee")) && ($q->param("date"))){

else{
	print "Usage: stalker.pl?stalkee=&lt;twitter_name&gt;&date=[mm-dd-yyyy || 7-letter weekday]";
	print "Example: stalker.pl?stalkee=tonysmartmeter&date=Thursday";
}

# Calculate averages 
sub predictWeekDay {

	my $stalkeeName = shift;
	my $printDate = shift;
	my $weekDay = shift;
	my $file = shift;
	my ($absoluteLow, $absoluteHigh, $average, $lowAverage, $predict, @allDayWatts, @allDayWattsTime, $lowEnergy, @googleURLs, @schedule);

	my $description = $stalkeeName . " is home and active";
	my $summary = $stalkeeName . " is currently home and active. Directions to their place: " . $googleDirections;

	@googleURLs = createMaps($stalkeeName);
	$absoluteLow = 999999999;
	$absoluteHigh = 1;
	$lowEnergy = 0;
	$weekDay = substr($weekDay, 0, 3);

	tie my @fileLines, 'Tie::File', $file;

	# Keys are time in 15 minute intervals: 00:00, 00:15, ... 23:45
	my @timeWatts = ("0)/1") x 96;
	my @tempTZ = parseCSV($fileLines[0]);
	my $tzoffset = $tempTZ[scalar(@tempTZ) - 2];

	# Read through each line of the data file looking for the specific day
	foreach my $line (@fileLines){
		my @thisTweet = parseTweets($line);
		my $tweetTime = Time::Piece->strptime($thisTweet[1], "%a %b %d %H:%M:%S %z %Y");

		$tweetTime += $tzoffset;
		my $tweetDayTemp = $tweetTime->wdayname();
		# if the input day matches that particular tweets day
		if ($weekDay =~ m/^$tweetDayTemp$/i){
			
			# Set the highs and lows
			if ($thisTweet[0] > $absoluteHigh){
				$absoluteHigh = $thisTweet[0];
			}
			if (($thisTweet[0] < $absoluteLow) && ($thisTweet[0] > 0)){
				$absoluteLow = $thisTweet[0];
			}
			push(@allDayWatts, $thisTweet[0]);
			push(@allDayWattsTime, $tweetTime);

		} # Ends if ($weekDay =~ m/^$tweetDayTemp$/i){

	} # Ends foreach my $line (@fileLines){

	untie @fileLines;

	$average = ($absoluteHigh + $absoluteLow) / 2;

	# Find values < average
	my $count = 0;
	for my $w (@allDayWatts){

		if ($w < $average){
			$lowEnergy += $w;
			$count += 1;
		}
	}

	$lowAverage = $lowEnergy / $count;

	my $counter = 0;
	# Calculate the averages for each 15 minutes of the day and store them in @timeWatts
	while ($counter < scalar(@allDayWatts)){
		
		my $timeTemp = $allDayWattsTime[$counter];
		my $hourIndex = $timeTemp->hour();
		my $minuteIndex = $timeTemp->min();
		my $arrayIndex = $hourIndex * 4;

		if (($minuteIndex > 15) && ($minuteIndex < 31)){
			$arrayIndex += 1;
		}
		elsif (($minuteIndex > 30) && ($minuteIndex < 46)){
			$arrayIndex += 2;
		}
		elsif (($minuteIndex > 45) && ($minuteIndex < 60)){
			$arrayIndex += 3;
		}

		my $tempDivisor = substr($timeWatts[$arrayIndex], -1);
		$tempDivisor += 1;
		$timeWatts[$arrayIndex] = $allDayWatts[$counter] . " + " . substr($timeWatts[$arrayIndex], 0, -1) . $tempDivisor; 

		$counter += 1;
	}

	
	$hourCount = 0;
	@watts = (0..95);
	# Evaulate the averages for each quarter hour
	while ($hourCount < scalar(@timeWatts)){
		my $tempDivisor2 = substr($timeWatts[$hourCount], -1);

		if ($tempDivisor2 > 1){
			$tempDivisor2 -= 1;
		}
		my $thisHour = "(" . substr($timeWatts[$hourCount], 0, -1) . $tempDivisor2;
		$watts[$hourCount] = eval($thisHour);
		$hourCount += 1;
	}

	my $timeIndex = Time::Piece->strptime("00:00:00", "%H:%M:%S");
	my $timeBegin;
	my $wattIndex = 0;
	my $beginBreakin = 0;
	print "On ", HTML::Entities::encode($printDate), ", that girl you call your girlfriend (", HTML::Entities::encode($stalkeeName), ") will most likely be home and awake during: ";
	while ($wattIndex < scalar(@watts)){
		
		# If the predicted wattage is less than the average of the low wattage + an error rate,
		# then we are predicting that they are either not home or sleeping
		if ($watts[$wattIndex] > ($lowAverage + $ERROR)){
		
			if ($beginBreakin == 0){
				$timeBegin = $timeIndex;
				$beginBreakin += 1;
			}

			elsif (($beginBreakin % 2) == 0){
				$timeBegin = $timeIndex;
				$beginBreakin += 1;
			}
		}

		elsif (($watts[$wattIndex] <= ($lowAverage + $ERROR)) && (($beginBreakin % 2) == 1)){

			# Subtracting two Time::Peice's will result in a Time::Second object
			# Give the burglar at least an hour to breakin, otherwise there just isn't time
			if ($beginBreakin == 1){
				my $printHour = $timeBegin->hour();
				if ($timeBegin->hour() < 10){
					$printHour = "0" . $timeBegin->hour();
				}
				my $printMin = $timeBegin->min();
				if ($timeBegin->min() < 10){
					$printMin = "0" . $timeBegin->min();
				}
				my $printEndHour = $timeIndex->hour();
				if ($timeIndex->hour() < 10){
					$printEndHour = "0" . $timeIndex->hour();
				}
				my $printEndMin = $timeIndex->min;
				if ($timeIndex->min() < 10){
					$printEndMin = "0" . $timeIndex->min();
				}

				my $icalFile = createICal($stalkeeName, $printDate, $printHour, $printMin, $printEndHour, $printEndMin, $googleURLs[0], $googleURLs[1], $description, $summary);
				print "<p><a href=\"", $icalFile, "\">", $printHour, ":", $printMin, " - ";
				print $printEndHour, ":", $printEndMin, "</a></p>";
				push(@schedule, $printHour . ":" . $printMin . "-" . $printEndHour . ":" . $printEndMin);
			}
			else{
				my $printHour2 = $timeBegin->hour();
				if ($timeBegin->hour() < 10){
					$printHour2 = "0" . $timeBegin->hour();
				}
				my $printMin2 = $timeBegin->min();
				if ($timeBegin->min() < 10){
					$printMin2 = "0" . $timeBegin->min();
				}
				my $printEndHour2 = $timeIndex->hour();
				if ($timeIndex->hour() < 10){
					$printEndHour2 = "0" . $timeIndex->hour();
				}
				my $printEndMin2 = $timeIndex->min();
				if ($timeIndex->min() < 10){
					$printEndMin2 = "0" . $timeIndex->min();
				}

				my $icalFile = createICal($stalkeeName, $printDate, $printHour2, $printMin2, $printEndHour2, $printEndMin2, $googleURLs[0], $googleURLs[1], $description, $summary);				
print "<p><a href=\"", $icalFile, "\">", $printHour2, ":", $printMin2, " - ";
				print $printEndHour2, ":", $printEndMin2, "</a></p>";

				push(@schedule, $printHour2 . ":" . $printMin2 . "-" . $printEndHour2 . ":" . $printEndMin2);
			}
			$beginBreakin += 1;
		}

		$timeIndex += (15 * 60);
		$wattIndex += 1;
	}

	printSchedule($stalkeeName, $printDate, $googleURLs[0], $googleURLs[1], @schedule);
	print "<p></p><p></p><h3>Happy Hunting!</h3>";


# Hardcoded in for demo, remove all lines except for printMaps
if ($stalkeeName eq "socawatt"){
printSocawattMaps();
}
else{

	printMaps($googleURLs[0], $googleURLs[1]);

}

} # Ends predictWeekDay()


# Hardcoded for no-internet access demo
sub printSocawattMaps{

	print "<a href=\"/location.png\">Location</a>";
	print "<p><a href=\"/directions.png\">Directions</a></p>";


}

sub printSchedule{

	my ($stalkeeName, $printDate, $googleLocation, $googleDest, @sched) = @_;

	my @times = split("-", $sched[0]);
	my @time1 = split(":", $times[0]);
	my @time2 = split(":", $times[1]);
	my @times2 = split("-", $sched[1]);
	my @times2HM = split(":", $times2[0]);

	my $timeZero = Time::Piece->strptime("00:00:00", "%H:%M:%S");
	my $timeWake = Time::Piece->strptime("10:00:00", "%H:%M:%S");
	my $initTime = Time::Piece->strptime($times[0], "%H:%M");
	my $endInitTime = Time::Piece->strptime($times[1], "%H:%M");
	my $secondBeginTime = Time::Piece->strptime($times2[0], "%H:%M");

	my $description = $stalkeeName . " is sleeping";
	my $summary = $stalkeeName . " is currently sleeping. Directions to their place: " . $googleDirections;

	# If the initial time is not midnight, then they are asleep
	if (($initTime > $timeZero) && (($initTime - $timeZero) > 14400) && ($initTime < $timeWake) ){
		print "<p>Also, they will most likely be sleeping between: ";
		my $icalFile = createICal($stalkeeName, $printDate, "00", "00", $time1[0], $time1[1], $googleLocation, $googleDest, $description, $summary);
		print "<p><a href=\"", $icalFile, "\">", "00:00", " - " . $times[0], "</a></p>";
	}

	# If the initial
	elsif (($initTime == $timeZero) && (($secondBeginTime - $endInitTime) > 14400) && ($secondBeginTime < $timeWake)){
		print "<p>Also, they will most likely be sleeping between: ";
		my $icalFile = createICal($stalkeeName, $printDate, $time2[0], $time2[1], $times2HM[0], $times2HM[1], $googleLocation, $googleDest, $description, $summary);
		print "<p><a href=\"", $icalFile, "\">", $times[1], " - " . $times2[0], "</a></p>";
	}

} # Ends sub printSchedule

# createICal($stalkeeName, $printDate, $printHour, $printMin, $printEndHour, $printEndMin, $googleURLs[0], $googleURLs[1]);
sub createICal{

	my $name = shift;
	my $date = shift;
	my $beginHour = shift;
	my $beginMin = shift;
	my $endHour = shift;
	my $endMin = shift;
	my $googleLocation = shift;
	my $googleDirections = shift;
	my $summary = shift;
	my $description = shift;

	my $calendar = Data::ICal->new();
	my $vevent = Data::ICal::Entry::Event->new();


	my $startTime;
	my $endTime;
	my @dateElements;

	# If they want to know based on a certain day of the week
	if ($date =~ m/s$/i){
		
		my $inDate = "next " . substr($date, 0, length($date) - 1);
		my $nextOccurrence = UnixDate(ParseDate($inDate), "%m-%d-%Y");
		@dateElements = split("-", $nextOccurrence);
		
	}

	# If they specified a specific date, $date = mm-dd-yyyy
	else{
		@dateElements = split("-", $date);
	}

	$startTime = DateTime->new(year => $dateElements[2], month => $dateElements[0], day => $dateElements[1], hour => $beginHour, minute => $beginMin, second => 0);
	$endTime = DateTime->new(year => $dateElements[2], month => $dateElements[0], day => $dateElements[1], hour => $endHour, minute => $endMin, second => 0);
	$vevent->add_properties(
		summary => $summary,
		description => $description . $googleDirections,
		location => $googleLocation
	);


	$vevent->start($startTime);
	$vevent->end($endTime);

	$calendar->add_entry($vevent);

	my $fileLocation = "/CalendarFiles/" . $name . "-" . $date . "-" . $beginHour . $beginMin . "-" . $endHour . $endMin . ".ics";

	open (OFILE, ">.." . $fileLocation);
	print OFILE $calendar->as_string;
	close (OFILE);

	return $fileLocation;

} # Ends createICal


# Plots a google map
sub createMaps{

	$name = shift;
	open(IFILE, "DataFiles/" . $name . ".dat");

	my $line;
	my $longitude;
	my $latitude;
	while ($line = <IFILE>){
		chomp($line);
		if ($line =~ m/location/i){
			my @coords = split("\'", $line);
			my @realCoords = split(/ /, $coords[3]);
			$longitude = $realCoords[0];
			$latitude = $realCoords[1];
			last;
		}
	}
	close(IFILE);

	my $googleLocationURL = "http://maps.google.com/maps?q=" . $longitude . "+" . $latitude . "+" . "(" . $name . ")" . "&iwloc=A&hl=en";
	my $googleDirections = "http://maps.google.com/?saddr=" . $CURRENTADDRESS . "&daddr=" . $longitude . "+" . $latitude . "(" . $name . ")";
	my @urls = ($googleLocationURL, $googleDirections);
	return @urls;

} # Ends sub createMaps{


# Prints out the map links
sub printMaps{
	my $location = shift;
	my $directions = shift;

	print "<p><a href=\"", $location, "\">Location</a></p>";
	print "<p><a href=\"", $directions, "\">Directions</a></p>";
}


# Parse a string that holds a single line from the stalkee data file
# Should return an array of the watts and date
# parsedTweet[0] = <watts>
# parsedTweets[1] = <time>
sub parseTweets {

	my $inText = shift;
	my @parsedTweet = ();
	my @parsedLine = parseCSV($inText);

	# Grab the Watt amounts
	my @tempWattage;
	if ($parsedLine[1] =~ m/Watt/){
		@tempWattage = split('Watt', $parsedLine[1]);
	}
	elsif ($parsedLine[1] =~ m/\sW,/){
		@tempWattage = split(/\sW/, $parsedLine[1]);
	}
	my @watt = split(' ', $tempWattage[0]);
	my $wattLength = scalar(@watt);
	push(@parsedTweet, $watt[$wattLength-1]);
	
	# Grab the date
	push(@parsedTweet, $parsedLine[2]);	

	return @parsedTweet;
}


# Parse a csv string into an array
sub parseCSV {
  my $text = shift;
  my @new = ();
  push (@new, $+) while $text =~ m{
          "([^\"\\]*(?:\\.[^\"\\]*)*)",?
          | ([^,]+),?
          | ,
      }gx;
  push (@new, undef) if substr($text, -1, 1) eq ",";
  return @new;
}

